#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Store conditions to trigger incidents.
#
class Incident < ApplicationRecord
  include Redborder::Audit
  include IncidentsHelper

  # ==========================================================
  # Constants
  # ==========================================================
  MULTIPLIER_RANGE = [0, 3, 6, 9]
  INTERVAL_RANGE = [300, 900, 1800, 3600, 21_600, 43_200, 86_400]
  TIME_WINDOW_RANGE = [ 300, 600, 900, 1800, 3600, 7200, 28800, 86400, 172800, 259200, 345600, 432000, 518400, 604800]

  # ==========================================================
  # Custom attributes
  # ==========================================================

  # ==========================================================
  # Relations
  # ==========================================================

  has_many :incident_users, dependent: :destroy  
  has_many :users, through: :incident_users
  has_many :notes, dependent: :destroy

  # Link / unlink incidents
  has_many :incident_links, dependent: :destroy
  has_many :linked_incidents, through: :incident_links, source: :linked_incident

  # Reverse relationship
  has_many :inverse_incident_links, class_name: 'IncidentLink', foreign_key: 'linked_incident_id', dependent: :destroy
  has_many :inverse_linked_incidents, through: :inverse_incident_links, source: :incident

  belongs_to :priority
  belongs_to :incident_status

  belongs_to :sensor

  has_one :playbook, dependent: :destroy
  has_one :incident_reports, dependent: :destroy

  # ================================================================
  # Callbacks
  # ================================================================

  after_create :assign_playbook
  after_create :create_incident_report

  # ================================================================
  # Validations
  # ================================================================

  validates :uuid, presence: true, uniqueness: true

  validates :name, presence: true,  length: { in: 1..255 }

  validates :priority, presence: true

  validates :incident_status, presence: true

  audited on: [:create, :update, :destroy]

  # ================================================================
  # Public methods
  # ================================================================

  def closed?
    incident_status.name == IncidentStatus::CLOSED
  end

  def closed!
    assign_status_by_name!(IncidentStatus::CLOSED)
  end

  def containment_achieved?
    incident_status.name == IncidentStatus::CONTAINMENT_ACHIEVED
  end

  def containment_achieved!
    assign_status_by_name!(IncidentStatus::CONTAINMENT_ACHIEVED)
  end

  def incident_reported?
    incident_status.name == IncidentStatus::INCIDENT_REPORTED
  end

  def incident_reported!
    assign_status_by_name!(IncidentStatus::INCIDENT_REPORTED)
  end

  def new?
    incident_status.name == IncidentStatus::NEW
  end

  def new!
    assign_status_by_name!(IncidentStatus::NEW)
  end

  def open?
    incident_status.name == IncidentStatus::OPEN
  end

  def open!
    assign_status_by_name!(IncidentStatus::OPEN)
  end

  def rejected?
    incident_status.name == IncidentStatus::REJECTED
  end

  def rejected!
    assign_status_by_name!(IncidentStatus::REJECTED)
  end

  def restoration_achieved?
    incident_status.name == IncidentStatus::RESTORATION_ACHIEVED
  end

  def restoration_achieved!
    assign_status_by_name!(IncidentStatus::RESTORATION_ACHIEVED)
  end

  def stalled?
    incident_status.name == IncidentStatus::STALLED
  end

  def stalled!
    assign_status_by_name!(IncidentStatus::STALLED)
  end

  # Combined view of all linked incidents
  def list_linked_incidents
    (linked_incidents + inverse_linked_incidents).uniq
  end

  # ================================================================
  # Public cache methods
  # ================================================================
  def self.cache_key_for_user(user_id)
    "rb-webui:incidents:#{user_id}"
  end
  
  # Save the current filters to Memcached
  def self.create_filters(user_id, search_params, name_param)
    filters = { search: search_params, name: name_param }
    Rails.cache.write(cache_key_for_user(user_id), filters, expires_in: 3.hours)
  end
  
  # Read the filters from Memcached
  def self.read_filters(user_id)
    Rails.cache.read(cache_key_for_user(user_id))
  end
  
  # Clear the filters from Memcached
  def self.clear_filters(user_id)
    Rails.cache.delete(cache_key_for_user(user_id))
  end

  # ================================================================
  # Private methods
  # ================================================================

  private

  def assign_status_by_name!(status_name)
    status = IncidentStatus.find_by(name: status_name)
    if status
      self.incident_status = status
      if save
        true
      else
        errors.full_messages.join(", ")
      end
    else
      false
    end
  end

  def assign_playbook
    assignment_rules = PlaybookAssignmentRule.where(enabled: true).order(:priority)
    playbook_assigned = false

    assignment_rules.each do |rule|
      next if playbook_assigned

      playbook_to_assign = PlaybookTemplate.find_by(id: rule.playbook_templates_id)

      if rule_conditions_met?(rule)
        playbook = create_playbook_from_template(playbook_to_assign)
        playbook_assigned = true
      end
    end
    assign_default_playbook unless playbook_assigned
  end

  def assign_default_playbook
    default_template = PlaybookTemplate.find_by(default: true)
    if default_template
      playbook = create_playbook_from_template(default_template)
      self.playbook = playbook
    end
  end

  def create_playbook_from_template(template)
    playbook = Playbook.create(name: template.name, description: template.description, incident: self)

    template.phase_templates.each do |phase_template|
      phase = playbook.phases.create(name: phase_template.name)

      phase_template.task_templates.each do |task_template|
        phase.tasks.create(name: task_template.name, summary: task_template.summary, description: task_template.description, execution: task_template.execution, integration_service: task_template.integration_service, action: task_template.action, position: task_template.position)
      end
    end

    playbook
  end

  def create_incident_report
    IncidentReports.create(incident: self)
  end

  def rule_conditions_met?(assignment_rule)
    return true if assignment_rule.playbook_assignment_rules_conditions.empty?

    execute_assignment_rule_conditions(assignment_rule)
  end

  def execute_assignment_rule_conditions(assignment_rule)
    if assignment_rule.condition_type == 'AND'
      assignment_rule.playbook_assignment_rules_conditions.all? { |condition| execute_condition(condition) }
    else
      assignment_rule.playbook_assignment_rules_conditions.any? { |condition| execute_condition(condition) }
    end
  end

  def execute_condition(condition)
    sub_property_value = if condition.property == 'Incidents'
                           send(condition.sub_property).to_s
                         elsif condition.property == 'Mitre'
                           mitre_data = get_incident_mitre_data(self)
                           if condition.sub_property == 'name' || condition.sub_property.include?('tactic')
                             mitre_data[:tactics]
                           elsif condition.sub_property.include?('subtechnique')
                             mitre_data[:subtechniques]
                           else
                             mitre_data[:techniques]
                           end
                         end

    condition.execute(sub_property_value)
  end
end
